
/** All functions below are modified from native Fx Customize code 
 *  in ../toolkit/customizeToolbar.js, v1.43.  The major changes are to wrap 
 *  functions in a for(){} loop to handle multiple toolboxes, to change the 
 *  reference from the single gToolbox to multiple toolbox elements within the 
 *  loop, support for statusbar as a toolbar. 
 */

InitWithToolbox = function (aToolbox) {

  gToolbox = aToolbox;
  gToolboxDocument = gToolbox.ownerDocument;

  // Pass the full array of toolboxes as the 2nd argument so as not to 
  // break any other xul app (extension) using customisable toolbars
  // Note: statusbar is both a 'toolbox' and a 'toolbar' here..

  gToolboxes = (window.arguments && window.arguments.length > 1) ? 
      window.arguments[1] : [gToolbox];

  gStatusbarMode = gToolboxes[0].localName == "statusbar" ? true : false;
  var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
                              .getService(Components.interfaces.nsIWindowMediator);
  var win = wm.getMostRecentWindow("navigator:browser");
  if(win) {
    gStatusbar = win.document.getElementById("status-bar");
    gNavToolbox = win.getNavToolbox();
  }
  for(var i = 0; i < gToolboxes.length; i++) {
    gToolboxes[i].addEventListener("draggesture", onToolbarDragGesture, false);
    gToolboxes[i].addEventListener("dragover", onToolbarDragOver, false);
    gToolboxes[i].addEventListener("dragexit", onToolbarDragExit, false);
    gToolboxes[i].addEventListener("dragdrop", onToolbarDragDrop, false);
    // For css styling of min-width etc.
    gToolboxes[i].setAttribute("incustomisemode", true);
  }

  if (gStatusbarMode) {
    gToolbox = gStatusbar;
    var dd = gStatusbar.getAttribute("ondragdrop");
    if (dd) {
      gStatusbar.setAttribute("tt_ondragdrop", dd);
      gStatusbar.removeAttribute("ondragdrop");
    }
  }

  initDialog();
}

function initDialog()
{
  // Move box into groupbox..
  var globalBox = document.getElementById("smallicons").parentNode;
  document.getElementById("tt-globalGroup").appendChild(globalBox);
  document.getElementById("smallicons").setAttribute("flex", 1);

  if (!gStatusbarMode) {
    var mode = gToolbox.getAttribute("mode");
    document.getElementById("modelist").value = mode;
    gToolboxIconSize = gToolbox.getAttribute("iconsize");
    var smallIconsCheckbox = document.getElementById("smallicons");
    smallIconsCheckbox.checked = gToolboxIconSize == "small";
    if (mode == "text")
      smallIconsCheckbox.disabled = true;

    // Hide global groupbox description for other apps (Console2 eg)
    if(window.opener.top.windowType != "navigator:browser")
      document.getElementById("tt-globalDesc").setAttribute("hidden", true);

    // Unset fixed width for vertical toolboxes while in Customize
    if(window.opener && window.opener.top && 
        typeof window.opener.top.tt_adjustToolboxWidth == "function") 
      window.opener.top.tt_adjustToolboxWidth(false);

  }
  else {
    var stringBundle = document.getElementById("totaltoolbarBundle");
    var title = stringBundle.getString("dialog.title.statusbar");
    document.title = title;
    var instructions = stringBundle.getString("instructions.description.statusbar");
    document.getElementById("instructions").setAttribute("value", instructions);

    document.getElementById("tt-globalGroupDeck").setAttribute("selectedIndex", 1);
  }
  
  // Build up the palette of other items.
  buildPalette();
  
  // Wrap all the items on the toolbar in toolbarpaletteitems.
  wrapToolbarItems();
}

repositionDialog = function () {
  // Always use persisted dimensions and position!
  return; 
}

removeToolboxListeners = function () {
  for(var i = 0; i < gToolboxes.length; i++) {
    gToolboxes[i].removeEventListener("draggesture", onToolbarDragGesture, false);
    gToolboxes[i].removeEventListener("dragover", onToolbarDragOver, false);
    gToolboxes[i].removeEventListener("dragexit", onToolbarDragExit, false);
    gToolboxes[i].removeEventListener("dragdrop", onToolbarDragDrop, false);
    gToolboxes[i].removeAttribute("incustomisemode");
  }
}

/**
 * Invoke a callback on the toolbox to notify it that the dialog is done 
 * and going away.
 */
notifyParentComplete = function () {
  for(var i = 0; i < gToolboxes.length; i++) {
    if("customizeDone" in gToolboxes[i]) gToolboxes[i].customizeDone(gToolboxChanged);
  }
}

/**
 * Persist the current set of buttons in all customizable toolbars to
 * localstore.
 */
//TODO: make reusable function 
function persistCurrentSets() {
  if (!gToolboxChanged)
  return;

  docElt = gToolboxDocument.documentElement;
  
  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = gToolboxes[j];
    var len = (gStatusbarMode) ? gToolboxes.length : gToolboxes[j].childNodes.length;

    var customCount = 0;
    var toolboxEmpty = true;
    for(var i = 0; i < len; ++i) {
      // Look for customizable toolbars that need to be persisted.
      var toolbar = (gStatusbarMode) ? 
              toolbox : (toolbox.childNodes ? toolbox.childNodes[i] : null);
      if (isCustomizableToolbar(toolbar)) {
        // Calculate currentset and store it in the attribute.
        var currentSet = toolbar.currentSet;
        toolbar.setAttribute("currentset", currentSet);

        var customIndex = toolbar.hasAttribute("customindex");
        if (customIndex) {
          if (!toolbar.firstChild) {
            // Remove any options from root doc and
            handleOptions("remove", toolbar);
            // Remove custom toolbars whose contents have been removed.
            toolbox.removeChild(toolbar);
            --i;
          } 
          else {
            toolboxEmpty = false;
            if(toolbox.toolbarset.hasAttribute('anonymous')) {
              if (currentSet != "__empty") {
                // Must persist xbl 'below-tabs' toolbar on the document root
                var attrPrefix = '_toolbarset.' + toolbox.toolbarset.getAttribute('id') + '.toolbar';
                docElt.setAttribute(attrPrefix+(++customCount), toolbar.toolbarName + ":" + currentSet);
                toolbar.setAttribute("customindex", customCount); //adjust count
                gToolboxDocument.persist(docElt.id, attrPrefix+customCount);
              }
            } 
            else {
              // Persist custom toolbar info on the <toolbarset/>
              toolbox.toolbarset.setAttribute("toolbar"+(++customCount), 
                  toolbar.toolbarName + ":" + currentSet);
              toolbar.setAttribute("customindex", customCount); //adjust count
              gToolboxDocument.persist(toolbox.toolbarset.id, "toolbar"+customCount);
            }
          }
        }

        if (!customIndex) {
          if (!toolbar.firstChild)
            // Remove any options from root doc
            handleOptions("remove", toolbar);
          else
            toolboxEmpty = false;
          if (currentSet != "__empty" || toolbar.hasAttribute("defaultset")) {
            if(toolbar.hasAttribute('anonymous')) {
              // Must persist xbl tabbar toolbars on the document root
              var attr = "_toolbar."+toolbar.getAttribute('id')+".currentset";
              docElt.setAttribute(attr, currentSet);
              gToolboxDocument.persist(gToolboxDocument.documentElement.id, attr);
            }
          }
          else
            if (!toolbar.hasAttribute("defaultset"))
              toolbar.removeAttribute("currentset");
        // Persist the currentset attribute directly on hardcoded toolbars.
        gToolboxDocument.persist(toolbar.id, "currentset");
        }
      }
    }
    // Number of remaining toolbars in toolbox
    toolbox.customToolbarCount = customCount; 
    setAttribute(toolbox, "empty", toolboxEmpty);
    gToolboxDocument.persist(toolbox.id, "empty");

    // Remove toolbarX attributes from toolbarset for removed toolbars, ie any 
    // entries past those for toolbars already indexed..
    if(toolbox.toolbarset) {
      if(toolbox.toolbarset.hasAttribute('anonymous')) {
        var attrPrefix = '_toolbarset.' + toolbox.toolbarset.getAttribute('id') + '.toolbar';
        while (docElt.hasAttribute(attrPrefix+(++customCount))) {
          docElt.removeAttribute(attrPrefix+customCount);
          gToolboxDocument.persist(docElt.id, attrPrefix+customCount);
        }
      }
      else {
        while (toolbox.toolbarset.hasAttribute("toolbar"+(++customCount))) {
          toolbox.toolbarset.removeAttribute("toolbar"+customCount);
          gToolboxDocument.persist(toolbox.toolbarset.id, "toolbar"+customCount);
        }
      }
    }
  }
  // Set fixed width for vertical toolboxes, for sidebar slider to work right
  if(window.opener && window.opener.top && 
      typeof window.opener.top.tt_adjustToolboxWidth == "function") 
    window.opener.top.tt_adjustToolboxWidth(true);
}

/**
 * Wraps all items in all customizable toolbars in a toolbox.
 */
wrapToolbarItems = function () {

  if (gStatusbarMode)
    window.opener.top.document.getElementById("tt-statusbar-spacer").setAttribute("flex", 1);

  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = (gStatusbarMode) ? gToolboxes : gToolboxes[j];
    var len = (gStatusbarMode) ? gToolboxes.length : gToolboxes[j].childNodes.length;

    for(var i = 0; i < len; ++i) {
      var toolbar = (gStatusbarMode) ? 
              toolbox[i] : (toolbox.childNodes ? toolbox.childNodes[i] : null);
      if(isCustomizableToolbar(toolbar)) {
        for(var k = 0; k < toolbar.childNodes.length; ++k) {
          var item = toolbar.childNodes[k];
          if(isToolbarItem(item)) {
            var nextSibling = item.nextSibling;

            var wrapper = wrapToolbarItem(item);

            if(nextSibling)
              toolbar.insertBefore(wrapper, nextSibling);
            else
              toolbar.appendChild(wrapper);

            cleanupItemForToolbar(item, wrapper);
          }
        }
      }
    }
  }
}

/**
  * Unwraps all items in all customizable toolbars in a toolbox.
  */
unwrapToolbarItems = function () {
  for(var j = 0; j < this.gToolboxes.length; j++) {
    var toolbox = gToolboxes[j];

    var paletteItems = toolbox.getElementsByTagName("toolbarpaletteitem");
    var paletteItem;
    while ((paletteItem = paletteItems.item(0)) != null) {
      var toolbarItem = paletteItem.firstChild;

      if (paletteItem.hasAttribute("itemdisabled"))
        toolbarItem.disabled = true;

      if (paletteItem.hasAttribute("itemhidden"))
        toolbarItem.hidden = true;

      if (paletteItem.hasAttribute("itemcollapsed"))
        toolbarItem.collapsed = true;

      // Remove label child from statusbarpanel item, if added..
      if (paletteItem.hasAttribute("itemlabelled")) {
        if (toolbarItem.getAttribute("class") == "statusbarpanel-iconic-text")
          toolbarItem.removeAttribute("label");
        else
          toolbarItem.removeChild(toolbarItem.lastChild);
      }

      // Restore style if any, remove forced display
      toolbarItem.removeAttribute("style");
      if (paletteItem.hasAttribute("itemstyle"))
        toolbarItem.setAttribute("style", paletteItem.getAttribute("itemstyle"));

      if (paletteItem.hasAttribute("itemcommand"))
        toolbarItem.setAttribute("command", paletteItem.getAttribute("itemcommand"));

      paletteItem.parentNode.replaceChild(toolbarItem, paletteItem);
    }
  }
  if (gStatusbarMode)
    window.opener.top.document.getElementById("tt-statusbar-spacer").setAttribute("flex", 0);
}

/**
 * Wraps an item that is currently on a toolbar and replaces the item
 * with the wrapper. This is not used when dropping items from the palette,
 * only when first starting the dialog and wrapping everything on the toolbars.
 */
function wrapToolbarItem(aToolbarItem)
{
  var wrapper = createWrapper(aToolbarItem.id);
  gToolboxDocument.adoptNode(wrapper);

  wrapper.flex = aToolbarItem.localName == "toolbarspring" ? 1 : 0;

  if (aToolbarItem.parentNode)
    aToolbarItem.parentNode.removeChild(aToolbarItem);
  
  wrapper.appendChild(aToolbarItem);
  
  return wrapper;
}

/**
 * Get the list of ids for the current set of items on each toolbar.
 */
getCurrentItemIds = function () {
  var currentItems = {};
  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = (gStatusbarMode) ? gToolboxes : gToolboxes[j];
    var len = (gStatusbarMode) ? gToolboxes.length : gToolboxes[j].childNodes.length;

    for (var i = 0; i < len; ++i) {
      var toolbar = (gStatusbarMode) ? 
          toolbox[i] : (toolbox.childNodes ? toolbox.childNodes[i] : null);
      if (isCustomizableToolbar(toolbar)) {
        var child = toolbar.firstChild;
        while (child) {
          if (isToolbarItem(child) && !child.getAttribute("removed"))
            currentItems[child.id] = 1;
          child = child.nextSibling;
        }
      }
    }
  }
  return currentItems;
}

/**
 * Makes sure that an item that has been cloned from a template
 * is stripped of all properties that may adversely affect it's
 * appearance in the palette.
 */
function cleanUpItemForPalette(aItem, aWrapper)
{
  aWrapper.setAttribute("place", "palette");
  setWrapperType(aItem, aWrapper);

  if (aItem.hasAttribute("title"))
    aWrapper.setAttribute("title", aItem.getAttribute("title"));
  else {
    if (isSpecialItem(aItem)) {
      var stringBundle = document.getElementById("stringBundle");
      var title = stringBundle.getString(aItem.id + "Title");
      aWrapper.setAttribute("title", title);
    }
    else if (!(aItem.hasAttribute("label") && (aItem.getAttribute("label") != ""))) {
      if (aItem.tooltipText)
        aWrapper.setAttribute("title", aItem.tooltipText);
      else
        aWrapper.setAttribute("title", aItem.id);
    }
  }
  // Remove attributes that screw up our appearance.
  aItem.removeAttribute("command");
  aItem.removeAttribute("observes");
  aItem.removeAttribute("disabled");
  aItem.removeAttribute("type");
  aItem.removeAttribute("hidden");
  aItem.removeAttribute("collapsed");

  if (aItem.localName == "toolbaritem" && aItem.firstChild) {
    aItem.firstChild.removeAttribute("observes");

    // So the throbber doesn't throb in the dialog,
    // cute as that may be...
    aItem.firstChild.removeAttribute("busy");
  }
}

/**
 * Makes sure that an item that has been cloned from a template
 * is stripped of all properties that may adversely affect it's
 * appearance in the toolbar.  Store critical properties on the 
 * wrapper so they can be put back on the item when we're done.
 */
function cleanupItemForToolbar(aItem, aWrapper)
{
  setWrapperType(aItem, aWrapper);
  aWrapper.setAttribute("place", "toolbar");

  if (aItem.hasAttribute("command")) {
    aWrapper.setAttribute("itemcommand", aItem.getAttribute("command"));
    aItem.removeAttribute("command");
  }

  if (aItem.hidden) {
    aWrapper.setAttribute("itemhidden", true);
    aItem.removeAttribute("hidden");
  }

  if (aItem.collapsed) {
    aWrapper.setAttribute("itemcollapsed", true);
    aItem.removeAttribute("collapsed");
  }

  if (aItem.disabled) {
    aWrapper.setAttribute("itemdisabled", true);
    aItem.disabled = false;
  }

  if (aItem.minWidth > 0) {
    aWrapper.setAttribute("itemminwidth", aItem.minWidth);
    aItem.minWidth = 0;
  }
  
  if (aItem.hasAttribute("removed")) {
    aWrapper.setAttribute("itemremoved", aItem.getAttribute("removed"));
  }
  
  // If no width (must allow for border etc incustomizemode), then no label 
  // or icon, so create label and insert id to identify the thing..
  if (aItem.boxObject.width <= 4 && !isSpecialItem(aItem)) {
    var id = aItem.id;
    // Add a label to statusbarpanel, for difficult to id statusbar items
    if (aItem.getAttribute("class") == "statusbarpanel-iconic-text" &&
        !aItem.hasAttribute("label"))
      aItem.setAttribute("label", aItem.hasAttribute("title") ? 
          aItem.getAttribute("title") : aItem.id);
    else 
      if (!aItem.hasAttribute("label")) {
        var label = document.createElement("label");
        gToolboxDocument.adoptNode(label);
        aItem.appendChild(label);
        aItem.lastChild.setAttribute("value", 
            aItem.hasAttribute("label") ? aItem.getAttribute("label") : 
            (aItem.hasAttribute("title") ? aItem.getAttribute("title") : 
            aItem.id));
        aWrapper.setAttribute("itemlabelled", true);
      }
  }

  // Force any "display: none" items to show up..  Set after boxObject check.
  if (aItem.hasAttribute("style"))
    aWrapper.setAttribute("itemstyle", aItem.getAttribute("style"));
  if (!aItem.hasAttribute("removed"))
    aItem.setAttribute("style", "display: -moz-box !important");
}

addNewToolbar = function () {
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                                .getService(Components.interfaces.nsIPromptService);

  var stringBundle = document.getElementById("stringBundle");
  var message = stringBundle.getString("enterToolbarName");
  var title = stringBundle.getString("enterToolbarTitle");

  var name = {};
  
  // Quitting from the toolbar dialog while the new toolbar prompt is up
  // can cause things to become unresponsive on the Mac. Until dialog modality  
  // is fixed (395465), disable the "Done" button explicitly.
  var doneButton = document.getElementById("donebutton");
  if (doneButton)
    doneButton.disabled = true;

  while(true) {

    if(!promptService.prompt(window, title, message, name, null, {})) {
      if (doneButton)
        doneButton.disabled = false;
      return;
    }

    if (!name.value) {
      message = stringBundle.getFormattedString("enterToolbarBlank", [name.value]);
     continue;
    }
    //TODO: need better error msg string 
    if (name.value.match(/[^\s|^\w]/g)) {
       message = stringBundle.getFormattedString("enterToolbarBlank", [name.value]);
       continue;
    }

    var dupeFound = false;

    // Check for an existing toolbar with the same display name 
    for(var j = 0; j < gToolboxes.length; j++) {
      var toolbox = gToolboxes[j];

      for(var i = 0; i < toolbox.childNodes.length; ++i) {
        var toolbar = toolbox.childNodes[i];
        var toolbarName = toolbar.getAttribute("toolbarname");

        if (toolbarName == name.value &&
            toolbar.getAttribute("type") != "menubar" &&
            toolbar.nodeName == 'toolbar') {
          dupeFound = true;
          break;
        }
      }
    }

    if (!dupeFound)
      break;

    message = stringBundle.getFormattedString("enterToolbarDup", [name.value]);
  }

  gToolbox.appendCustomToolbar(name.value, "", true);

  toolboxChanged();
  
  if (doneButton)
    doneButton.disabled = false;

}

/**
 * Restore the default set of buttons to fixed toolbars,
 * remove all custom toolbars, and rebuild the palette.
 */
restoreDefaultSet = function () {

  // Save disabled/command states, because we're
  // going to recreate the wrappers and lose this
  var savedAttributes = saveItemAttributes(["itemdisabled", "itemcommand"]);

  docElt = gToolboxDocument.documentElement;

  for(var i = 0; i < gToolboxes.length; i++) {
    var toolbox = gToolboxes[i];

    if (!gStatusbarMode) {
      // Remove all of the customized toolbars.
      var child = toolbox.lastChild;
      while(child) {
        if(child.hasAttribute("customindex")) {
          // Clean up any customizations from the root doc
          handleOptions("remove", child); 
          var thisChild = child;
          child = child.previousSibling;
          toolbox.removeChild(thisChild);
        } else {
          child = child.previousSibling;
        }
      }
    }
    
    // Restore the default icon size and mode - meaning Fx main toolbox if none
    var defaultMode = gToolbox.hasAttribute("defaultmode") ? 
        gToolbox.getAttribute("defaultmode") : 
        (gNavToolbox ? gNavToolbox.getAttribute("defaultmode") : null);
    var defaultIconsSmall = gToolbox.hasAttribute("defaulticonsize") ? 
        (gToolbox.getAttribute("defaulticonsize") == "small" ? "small" : "large") :
        (gNavToolbox ? 
            (gNavToolbox.getAttribute("defaulticonsize") == "small" ? "small" : "large") :
            "small");

    setAttribute(toolbox, "iconsize", defaultIconsSmall);
    gToolboxDocument.persist(toolbox.id, "iconsize");
    document.getElementById("smallicons").checked = defaultIconsSmall == "small";
    document.getElementById("smallicons").disabled = (defaultMode == "text" ? true : false);
    setAttribute(toolbox, "mode", defaultMode);
    gToolboxDocument.persist(toolbox.id, "mode");
    document.getElementById("modelist").value = defaultMode;

    if (gStatusbarMode) {
      unwrapToolbarItems();
      tt_resetStatusbar();
    }
    else {
      // Restore the defaultset for fixed toolbars.
      var toolbar = toolbox.firstChild;
      while (toolbar) {
        if(isCustomizableToolbar(toolbar)) {
////          if(!toolbar.hasAttribute("customindex")) {
            var defaultSet = toolbar.getAttribute("defaultset");
            if(defaultSet || defaultSet == "")
              toolbar.currentSet = defaultSet;
            //Toolbox has custom binding, otherwise it's some other 3rd party toolbox
            if (typeof toolbar.parentNode.setToolbarIconSize == "function")
              toolbar.parentNode.setToolbarIconSize(toolbar, defaultIconsSmall);
            else 
              if (toolbar.getAttribute("currentset") != "__empty") {
                setAttribute(toolbar, "iconsize", defaultIconsSmall);
                gToolboxDocument.persist(toolbar.id, "iconsize");
              }
            // Toolbox has custom binding, otherwise it's some other 3rd party toolbox
            if (typeof toolbar.parentNode.setToolbarMode == "function")
              toolbar.parentNode.setToolbarMode(toolbar, defaultMode);
            else
              if (toolbar.getAttribute("currentset") != "__empty") {
                setAttribute(toolbar, "mode", defaultMode);
                gToolboxDocument.persist(toolbar.id, "mode");
              }
////          }
        }
        // Show nav-bar and bookmarks toolbar if hidden
        if (toolbar.id == "nav-bar" || toolbar.id == "PersonalToolbar")
          toolbar.removeAttribute("collapsed");

        toolbar = toolbar.nextSibling;
      }
    }
  }

  // Now rebuild the palette.
  buildPalette();

  // Now re-wrap the items on the toolbar.
  wrapToolbarItems();

  // Restore the disabled and command states
  restoreItemAttributes(["itemdisabled", "itemcommand"], savedAttributes);

  toolboxChanged();
}

updateIconSize = function (aUseSmallIcons, localDefault) {
  gToolboxIconSize = aUseSmallIcons ? "small" : "large";

  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = gToolboxes[j];

    setAttribute(toolbox, "iconsize", gToolboxIconSize);
    gToolboxDocument.persist(toolbox.id, "iconsize");

    for(var i = 0; i < toolbox.childNodes.length; ++i) {
      var toolbar = toolbox.childNodes[i];
      if(isCustomizableToolbar(toolbar)) {
        //Toolbox has custom binding, otherwise it's some other 3rd party toolbox
        if (typeof toolbar.parentNode.setToolbarIconSize == "function")
          toolbar.parentNode.setToolbarIconSize(toolbar, gToolboxIconSize);
        else {
          var toolbarIconSize = (localDefault && toolbar.hasAttribute("defaulticonsize")) ?
                                toolbar.getAttribute("defaulticonsize") :
                                gToolboxIconSize;
          setAttribute(toolbar, "iconsize", toolbarIconSize);
          gToolboxDocument.persist(toolbar.id, "iconsize");
        }
      }
    }
  }

  // Set fixed width for vertical toolboxes, for sidebar slider to work right
  if(window.opener && window.opener.top && 
      typeof window.opener.top.tt_adjustToolboxWidth == "function") 
    window.opener.top.tt_adjustToolboxWidth(true);
}

updateToolbarMode = function (aModeValue, localDefault) {

  for(var j = 0; j < gToolboxes.length; j++) {
    var toolbox = gToolboxes[j];

    setAttribute(toolbox, "mode", aModeValue);
    gToolboxDocument.persist(toolbox.id, "mode");

    for (var i = 0; i < toolbox.childNodes.length; ++i) {
      var toolbar = toolbox.childNodes[i];
      if (isCustomizableToolbar(toolbar)) {
        // Toolbox has custom binding, otherwise it's some other 3rd party toolbox
        if (typeof toolbar.parentNode.setToolbarMode == "function")
          toolbar.parentNode.setToolbarMode(toolbar, aModeValue);
        else {
          var toolbarMode = (localDefault && toolbar.hasAttribute("defaultmode")) ?
                            toolbar.getAttribute("defaultmode") :
                            aModeValue;
          setAttribute(toolbar, "mode", toolbarMode);
          gToolboxDocument.persist(toolbar.id, "mode");
        }
      }
    }
  }

  var iconSizeCheckbox = document.getElementById("smallicons");
  iconSizeCheckbox.disabled = aModeValue == "text";

  // Set fixed width for vertical toolboxes, for sidebar slider to work right
  if(window.opener && window.opener.top && 
      typeof window.opener.top.tt_adjustToolboxWidth == "function") 
    window.opener.top.tt_adjustToolboxWidth(true);
}

function isCustomizableToolbar(aElt)
{
  return (aElt) && ((aElt.localName == "toolbar" && aElt.getAttribute("customizable") ||
         aElt.localName == "statusbar"));
}

function isToolbarItem(aElt)
{
  return aElt.localName == "toolbarbutton" ||
         aElt.localName == "toolbaritem" ||
         aElt.localName == "toolbarseparator" ||
         aElt.localName == "toolbarspring" ||
         aElt.localName == "toolbarspacer" ||
         aElt.localName == "statusbarpanel";
}

///////////////////////////////////////////////////////////////////////////
//// Drag and Drop observers

var toolbarDNDObserver =
{
  onDragOver: function (aEvent, aFlavour, aDragSession)
  {
    var toolbar = aEvent.target;
    var dropTarget = aEvent.target;
    while (toolbar && (toolbar.localName != "toolbar" && toolbar.localName != "statusbar")) {
      dropTarget = toolbar;
      toolbar = toolbar.parentNode;
    }
    
    var previousDragItem = gCurrentDragOverItem;

    // Make sure we are dragging over a customizable toolbar.
    if (!isCustomizableToolbar(toolbar)) {
      gCurrentDragOverItem = null;
      return;
    }
    
    if (dropTarget.localName == "toolbar" || dropTarget.localName == "statusbar") {
      gCurrentDragOverItem = dropTarget;
    } else {
      gCurrentDragOverItem = null;

      var direction = window.getComputedStyle(dropTarget.parentNode, null).direction;
      var dropTargetCenter = dropTarget.boxObject.x + (dropTarget.boxObject.width / 2);
      if (direction == "ltr")
        dragAfter = aEvent.clientX > dropTargetCenter;
      else
        dragAfter = aEvent.clientX < dropTargetCenter;
        
      if (dragAfter) {
        gCurrentDragOverItem = dropTarget.nextSibling;
        if (!gCurrentDragOverItem)
          gCurrentDragOverItem = toolbar;
      } else
          gCurrentDragOverItem = dropTarget;
    }    

    if (previousDragItem && gCurrentDragOverItem != previousDragItem) {
      setDragActive(previousDragItem, false);
    }
    
    setDragActive(gCurrentDragOverItem, true);
    
    aDragSession.canDrop = true;
  },
  
  onDrop: function (aEvent, aXferData, aDragSession)
  {
    if (!gCurrentDragOverItem)
      return;
    
    setDragActive(gCurrentDragOverItem, false);

    var draggedItemId = aXferData.data;
    if (gCurrentDragOverItem.id == draggedItemId)
      return;

    var toolbar = aEvent.target;
    while (toolbar.localName != "toolbar" && toolbar.localName != "statusbar")
      toolbar = toolbar.parentNode;

    var draggedPaletteWrapper = document.getElementById("wrapper-"+draggedItemId);
    if (!draggedPaletteWrapper) {
      // The wrapper has been dragged from the toolbar.
      
      // Get the wrapper from the toolbar document and make sure that
      // it isn't being dropped on itself.
      var wrapper = gToolboxDocument.getElementById("wrapper-"+draggedItemId);
      if (wrapper == gCurrentDragOverItem)
        return;

      // Don't allow static kids (e.g., the menubar) to move.
      if (wrapper.parentNode.firstPermanentChild && wrapper.parentNode.firstPermanentChild.id == wrapper.firstChild.id)
        return;
      if (wrapper.parentNode.lastPermanentChild && wrapper.parentNode.lastPermanentChild.id == wrapper.firstChild.id)
        return;

      // Remove the item from it's place in the toolbar.
      wrapper.parentNode.removeChild(wrapper);

      // Determine which toolbar we are dropping on.
      var dropToolbar = null;
      if (gCurrentDragOverItem.localName == "toolbar" || 
          gCurrentDragOverItem.localName == "statusbar")
        dropToolbar = gCurrentDragOverItem;
      else
        dropToolbar = gCurrentDragOverItem.parentNode;
      
      // Insert the item into the toolbar.
      if (gCurrentDragOverItem != dropToolbar)
        dropToolbar.insertBefore(wrapper, gCurrentDragOverItem);
      else
        dropToolbar.appendChild(wrapper);
    } else {
      // The item has been dragged from the palette
      
      if (!gStatusbarMode || isSpecialItem(draggedPaletteWrapper.firstChild)) {
        // Create a new wrapper for the item. We don't know the id yet.
        var wrapper = createWrapper("");
        gToolboxDocument.adoptNode(wrapper);
  
        // Ask the toolbar to clone the item's template, place it inside the wrapper, and insert it in the toolbar.
        var newItem = toolbar.insertItem(draggedItemId, gCurrentDragOverItem == toolbar ? null : gCurrentDragOverItem, wrapper);
  
        // Prepare the item and wrapper to look good on the toolbar.
        cleanupItemForToolbar(newItem, wrapper);
        wrapper.id = "wrapper-"+newItem.id;
        wrapper.flex = newItem.flex;
      } 
      else {
        // Move the item in the statusbar.
        var wrapper = gToolboxDocument.getElementById("wrapper-"+draggedItemId);
        if (gCurrentDragOverItem != gStatusbar)
          gStatusbar.insertBefore(wrapper, gCurrentDragOverItem);
        else
          gStatusbar.appendChild(wrapper);
        wrapper.removeAttribute("itemremoved");
        wrapper.firstChild.removeAttribute("removed");
        wrapper.firstChild.setAttribute("style", "display: -moz-box !important");
      }

      // Remove the wrapper from the palette.
      var currentRow = draggedPaletteWrapper.parentNode;
      if (draggedItemId != "separator" &&
          draggedItemId != "spring" &&
          draggedItemId != "spacer") {
        currentRow.removeChild(draggedPaletteWrapper);

        while (currentRow) {
          // Pull the first child of the next row up
          // into this row.
          var nextRow = currentRow.nextSibling;
          
          if (!nextRow) {
            var last = currentRow.lastChild;
            var first = currentRow.firstChild;
            if (first == last) {
              // Kill the row.
              currentRow.parentNode.removeChild(currentRow);
              break;
            }

            if (last.localName == "spacer") {
              var flex = last.getAttribute("flex");
              last.setAttribute("flex", ++flex);
              // Reflow doesn't happen for some reason.  Trigger it with a hide/show. ICK! -dwh
              last.hidden = true;
              last.hidden = false;
              break;
            } 
            else {
              // Make a spacer and give it a flex of 1.
              var spacer = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                                    "spacer");
              spacer.setAttribute("flex", "1");
              currentRow.appendChild(spacer);
            }
            break;
          }
          
          currentRow.appendChild(nextRow.firstChild);
          currentRow = currentRow.nextSibling;
        }
      }
    }
    
    gCurrentDragOverItem = null;

    toolboxChanged();
  },
  
  _flavourSet: null,
  
  getSupportedFlavours: function ()
  {
    if (!this._flavourSet) {
      this._flavourSet = new FlavourSet();
      var documentId = gToolboxDocument.documentElement.id;
      this._flavourSet.appendFlavour("text/toolbarwrapper-id/"+documentId);
    }
    return this._flavourSet;
  }
}

var paletteDNDObserver =
{
  onDragOver: function (aEvent, aFlavour, aDragSession)
  {
    aDragSession.canDrop = true;
  },

  onDrop: function(aEvent, aXferData, aDragSession)
  {
    var itemId = aXferData.data;
    
    var wrapper = gToolboxDocument.getElementById("wrapper-"+itemId);
    if (wrapper && !wrapper.getAttribute("itemremoved")) {
      // Don't allow static kids (e.g., the menubar) to move.
      if (wrapper.parentNode.firstPermanentChild && wrapper.parentNode.firstPermanentChild.id == wrapper.firstChild.id)
        return;
      if (wrapper.parentNode.lastPermanentChild && wrapper.parentNode.lastPermanentChild.id == wrapper.firstChild.id)
        return;

      // The item was dragged out of the toolbar.
      if (gStatusbarMode && !isSpecialItem(wrapper.firstChild)) {
        wrapper.firstChild.setAttribute("removed", true);
        wrapper.setAttribute("itemremoved", true);
        wrapper.firstChild.removeAttribute("style");
      }
      else
        wrapper.parentNode.removeChild(wrapper);
      
      var wrapperType = wrapper.getAttribute("type");
      if (wrapperType != "separator" &&
          wrapperType != "spacer" &&
          wrapperType != "spring") {
        // Find the template node in the toolbox palette
        var templateNode = gToolbox.palette.firstChild;
        while (templateNode) {
          if (templateNode.id == itemId)
            break;
          templateNode = templateNode.nextSibling;
        }
        if (!templateNode)
          return;
        
        // Clone the template and add it to our palette.
        var paletteItem = templateNode.cloneNode(true);
        appendPaletteItem(paletteItem);
      }
    }

    toolboxChanged();
  },

  _flavourSet: null,

  getSupportedFlavours: function ()
  {
    if (!this._flavourSet) {
      this._flavourSet = new FlavourSet();
      var documentId = gToolboxDocument.documentElement.id;
      this._flavourSet.appendFlavour("text/toolbarwrapper-id/"+documentId);
    }
    return this._flavourSet;
  }
}
